package de.tub.mapc.Coordination;

import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

public class SwitchVotes{
	
	private List<SwitchVote> votes;
	private int votingCount;
	
	public SwitchVotes(){
		votes = new LinkedList<SwitchVote>();
	}
	
	public void addVoter(Voter voter, int dist){
		boolean voted = false;
		// check, wheter this voter already voted
		for(SwitchVote v : votes){
			if(v.voter == voter){
				v.valid = v.VOTE_VALID_FOR_STEPS;
				v.valid += voter.getValidSteps();
				v.distance = dist;
				voted = true;
				break;
			}
		}
		// if not, add new vote
		if(!voted)votes.add(new SwitchVote(voter, dist));
	}
	
	public void removeVoter(Voter voter){
		Iterator<SwitchVote> i = votes.iterator();
		while(i.hasNext()){
			SwitchVote v = i.next();
			if(v.voter == voter){
				i.remove();
				break;
			}
		}
	}
	
	public int getVotingCount(){
		if(votingCount == 0){
			for(SwitchVote v : votes){
				votingCount += v.voter.getPrio();
			}
		}
		return votingCount;
	}
	
	public List<Agent> getVotingAgents() {
		List<Agent> result = new LinkedList<Agent>();
		// sort the votes, so the nearest votes are first in the list an the agents of this vote
		// are added first to the result
		Collections.sort(votes, new CompareSwitchVotes());
		for(SwitchVote v : votes){
			result.addAll(v.voter.getAgents());
		}
		return result;
	}
	
	public void step(){
		Iterator<SwitchVote> i = votes.iterator();
		while(i.hasNext()){
			SwitchVote v = i.next();
			v.valid--;
			if (v.valid == 0) i.remove();
		}
		votingCount = 0;
	}
	
	/**
	 * 
	 * @return max distance between a voter and the fence
	 */
	public int getMaxDistance(){
		int dist=0;
		for(SwitchVote v: votes){
			if(v.getDistance()>dist)dist=v.getDistance();
		}
		return dist;
	}
	
	private class SwitchVote {
		
		private final int VOTE_VALID_FOR_STEPS = 3;
		
		private int valid;
		private Voter voter;
		private int distance;
		
		private SwitchVote(Voter v, int dist){
			valid = VOTE_VALID_FOR_STEPS;
			valid += v.getValidSteps();
			voter = v;
			distance = dist;
		}
		
		public int getDistance(){
			return distance;
		}
	}
	
	private class CompareSwitchVotes implements Comparator<SwitchVote>{
		
		@Override
		public int compare(SwitchVote a, SwitchVote b){
			//a.calculateCost();b.calculateCost();
			if (a.getDistance() == b.getDistance()) return 0;
			if (a.getDistance() > b.getDistance()) return 1;
			return -1;
		}
	}
	
	
}
