package de.tub.mapc.Coordination.goals;

import java.io.IOException;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.MessageService;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.pathfinding.PathResult;

public class GoalBlockEnemyCorral extends AbstractGoal {
	
	private Position goalPosition;
	private boolean isReached;
	
	public GoalBlockEnemyCorral(){
	}
	
	public GoalBlockEnemyCorral(Position blockingPosition) {
		goalPosition = blockingPosition;
	}

	@Override
	public Position getGoalPosition() {
		return goalPosition;
	}

	@Override
	public String getAction(Perception perception, Gridworld world,	IPathFindingHandle pathfinder, MessageService coordinatorMessageService) {
		try {
			// pathfinder befragen
			PathResult pr = pathfinder.getPathForBlocker(goalPosition, perception.getAgentsPosition(), world);
			if(pr.isReachedGoal()){
				
				// goal is reached, stay at hte switch until a new goal is received
				isReached = true;
				return this.getRandomAction();
				
			}else if(pr.getErrortype() != null){
							
				if (pr.getErrortype() == IMAPCConstants.PATH_ERROR){
					// there was an real error in the pathfinder
					// this should not happen
					// but when, the agent should move on
					// -> random move
//					_logger.error("PATH: THERE WAS AN ERROR");
					return this.getRandomAction();
				}else if (pr.getErrortype() == IMAPCConstants.PATH_NOTFOUND){
					// the pathfinder was not able to find a path
					// should not happen, because this goal is only assigned if there is an path 
					return this.getRandomAction();
				
				}else if (pr.getErrortype() == IMAPCConstants.PATH_FENCEAHEAD){
					// the pathfinder indicates there is a fence in the next move,
					// maybe we already voted but the switcher hasn't reached the switching position yet or we didn't vote
					Switch s = world.findSwitchForFence(new Position(pr.getFencex(),pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(), pr.getFencey()),true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
					// If the fence is in the next move, the action is skip
					return pr.getThisStep();
				}else if (pr.getErrortype() == IMAPCConstants.PATH_BLOCKED){
					// The Pathfinder indicates the next move is blocked
					// da is ne kuh oder nen agent ins bild gerannt, // TODO randommove oder lieber skip?
					return this.getRandomAction();
				}
			}else{
				// no error, vote if there are fences ahead
				if(pr.getStepsToNextFence() != -1){
					//TODO ab welchem Abstand anmelden?
					Position s = world.findSwitchForFence(new Position(pr.getFencex(), pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(),pr.getFencey()),true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
				}
				
				return pr.getThisStep();
			}
		}catch (Exception e){
			e.printStackTrace();
		}
		return this.getRandomAction();
	}

	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		goalPosition = (Position)arg0.readObject();
		this.isReached = arg0.readBoolean();
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		arg0.writeObject(goalPosition);
		arg0.writeBoolean(this.isReached);
	}

	@Override
	public String toString() {
		return "block enemy corral " + goalPosition;
	}

	@Override
	public int getPriority() {
		return 1;
	}

	@Override
	public boolean isAchieved() {
		return isReached;
	}

	@Override
	public boolean equals(Object o) {
		if(o == null){
			return false;
		}
		if(o instanceof GoalBlockEnemyCorral){
			GoalBlockEnemyCorral other = (GoalBlockEnemyCorral) o;
			if (other.goalPosition.equals(goalPosition)){
				return true;
			}
		}
		return false;
	}

}
