package de.tub.mapc.Coordination.goals;

import java.awt.Point;
import java.io.IOException;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.MessageService;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.pathfinding.PathResult;

public class GoalDriveHerd extends AbstractGoal {
	
	private Position herdDrivingPos;
	private Point upperLeft;
	private Point lowerRight;
	private boolean isReached;
	private String lastAction = IMAPCConstants.ACTION_SKIP;
	
	public GoalDriveHerd() {
		
	}
	
	public GoalDriveHerd(Position pos, Point ul, Point lr) {
		herdDrivingPos = pos;
		upperLeft = ul;
		lowerRight = lr;
	}

	public Point getUpperLeft(){
		return this.upperLeft;
	}
	
	public Point getLowerRight(){
		return this.lowerRight;
	}	
	
	
	@Override
	public String getAction(Perception perception, Gridworld world,
			IPathFindingHandle path, MessageService coordinatorMessageService) {
		Position target=herdDrivingPos;
		if(target==null)return IMAPCConstants.ACTION_SKIP;
		else{
			PathResult pr = path.getPathForDriver(target, perception.getAgentsPosition(), world, upperLeft, lowerRight);
			if(pr.isReachedGoal()){
				//TODO weiter in die letzte Richtung laufen?!?
				return lastAction; // statt this.getRandomAction();
			}
			else if(pr.getErrortype() != null){
				
				if (pr.getErrortype() == IMAPCConstants.PATH_ERROR){
					// there was an real error in the pathfinder
					// this should not happen
					// but when, the explorer should move on
					// -> random move
					return this.getRandomAction();
					
				}else if (pr.getErrortype() == IMAPCConstants.PATH_NOTFOUND){
					// The pathfinder was not able to find a path
					return this.getRandomAction();
					
				}else if (pr.getErrortype() == IMAPCConstants.PATH_FENCEAHEAD){
					// the pathfinder indicates there is a fence on the path
					// maybe we already voted but the switcher hasn't reached the switching position yet or we didn't vote
					Switch s = world.findSwitchForFence(new Position(pr.getFencex(),pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(), pr.getFencey()), true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
					// if the fence is in the next move, the action is skip
					return pr.getThisStep();
					
				}else if (pr.getErrortype() == IMAPCConstants.PATH_BLOCKED){
					// The Pathfinder indicates the next move is blocked
					// this happens, when another agent or a cow moved to this field
					// -> random action is better than no move
					return this.getRandomAction();
				}
			}else{
				// no error, vote if there are fences ahead
				if(pr.getStepsToNextFence() != -1){
					//TODO ab welchem Abstand anmelden?
					Position s = world.findSwitchForFence(new Position(pr.getFencex(), pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(), pr.getFencey()), true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
				}
				// TODO weitere Zauune anmelden
				//save action in lastAction
				lastAction = pr.getThisStep();
				
				// return action
				return pr.getThisStep();
			}
		}
		return null;
	}

	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		this.isReached=arg0.readBoolean();
		this.herdDrivingPos=(Position)arg0.readObject();
		int x = arg0.readInt();
		int y = arg0.readInt();
		upperLeft = new Point(x,y);
		x = arg0.readInt();
		y = arg0.readInt();
		lowerRight = new Point(x,y);
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		arg0.writeBoolean(isReached);
		arg0.writeObject(herdDrivingPos);
		arg0.writeInt(upperLeft.x);
		arg0.writeInt(upperLeft.y);
		arg0.writeInt(lowerRight.x);
		arg0.writeInt(lowerRight.y);
	}

	@Override
	public String toString() {
		return "drive herd to: "+ (herdDrivingPos == null ? "null" : herdDrivingPos.toString());
	}

	@Override
	public int getPriority() {
		return 40;
	}

	@Override
	public boolean isAchieved() {
		return false;
	}

	@Override
	public boolean equals(Object o) {
		if(o == null){
			return false;
		}
		if (herdDrivingPos == null) return false;
		if(o instanceof GoalDriveHerd){
			GoalDriveHerd other = (GoalDriveHerd) o;
			if (herdDrivingPos.equals(other.herdDrivingPos)){
				return true;
			}
		}
		return false;
	}

	@Override
	public Position getGoalPosition() {
		return herdDrivingPos;
	}
}
