package de.tub.mapc.Coordination.goals;

import java.io.IOException;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.MessageService;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.pathfinding.PathResult;

public class GoalExploreField extends AbstractGoal {

	private Position goalPosition;
	private boolean isReached;
	
	public GoalExploreField(){
		//needed for deserializer
	}
	
	public GoalExploreField(Position p) {
		goalPosition = p;
	}
	
	@Override
	public Position getGoalPosition(){
		return goalPosition;
	}
	
	@Override
	public String getAction(Perception perception, Gridworld world,
			IPathFindingHandle pathfinder, MessageService coordinatorMessageService) {
		
		// check, whether the field is already explored
		if(world.getGrid()[goalPosition.getX()][goalPosition.getY()].getType() != IMAPCConstants.TYPE_UNKNOWN){
			this.isReached = true;
			return this.getRandomAction();
		}
		
		//if not, explore the field
		try {
			// find path to the field
			PathResult result = pathfinder.getPath(goalPosition, perception.getAgentsPosition(),world, true);
			
			if(result.isReachedGoal()){
				isReached = true;
				// the goal is reached, explorer should always move, so make a random move
				return this.getRandomAction();
				
			}else if(result.getErrortype() != null){
				
				if (result.getErrortype() == IMAPCConstants.PATH_ERROR){
					// there was an real error in the pathfinder
					// this should not happen
					// but when, the explorer should move on
					// -> random move
					return this.getRandomAction();
					
				}else if (result.getErrortype() == IMAPCConstants.PATH_NOTFOUND){
					// The pathfinder was not able to find a path
					return this.getRandomAction();
					
				}else if (result.getErrortype() == IMAPCConstants.PATH_FENCEAHEAD){
					// the pathfinder indicates there is a fence on the path
					// maybe we already voted but the switcher hasn't reached the switching position yet or we didn't vote
					Switch s = world.findSwitchForFence(new Position(result.getFencex(),result.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(result.getFencex(), result.getFencey()), true, result.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, result.getStepsToNextFence());
					}
					// if the fence is in the next move, the action is skip
					return result.getThisStep();
					
				}else if (result.getErrortype() == IMAPCConstants.PATH_BLOCKED){
					// The Pathfinder indicates the next move is blocked
					// this happens, when another agent or a cow moved to this field
					// -> random action is better than no move
					return this.getRandomAction();
				}
				
			}else{

				// no error, vote if there are fences ahead
				// explorer needs only the next fence, thats enough
				if(result.getStepsToNextFence() != -1){
					//TODO ab welchem Abstand anmelden?
					Position s = world.findSwitchForFence(new Position(result.getFencex(), result.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(result.getFencex(), result.getFencey()), true, result.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, result.getStepsToNextFence());
					}
				}
				// return action
				return result.getThisStep();
			}
		}catch (Exception e){
			e.printStackTrace();
		}
		// we get here only if there was an exception
		// randommove is better than nothing
		return this.getRandomAction();
	}

	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		goalPosition = (Position)arg0.readObject();
		this.isReached = arg0.readBoolean();
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		arg0.writeObject(goalPosition);
		arg0.writeBoolean(this.isReached);
	}

	@Override
	public String toString() {
		return "explore field " + goalPosition;
	}

	@Override
	public int getPriority() {
		return 10;
	}

	@Override
	public boolean isAchieved() {
		return isReached;
	}

	@Override
	public boolean equals(Object o) {
		if(o == null){
			return false;
		}
		if(o instanceof GoalExploreField){
			GoalExploreField other = (GoalExploreField) o;
			if (other.goalPosition.equals(goalPosition)){
				return true;
			}
		}
		return false;
	}

}
