package de.tub.mapc.Coordination.goals;

import java.io.IOException;

import de.dailab.keks.io.IDeserialiser;
import de.dailab.keks.io.ISerialiser;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.components.contestagents.MessageService;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.knowledge.Switch;
import de.tub.mapc.pathfinding.PathResult;

public class GoalReachDrivingPosition extends AbstractGoal {

	private Position goalPosition;
	private boolean isReached;
	
	public GoalReachDrivingPosition(Position p) {
		goalPosition = p;
	}
	
	public GoalReachDrivingPosition(int x, int y) {
		goalPosition = new Position(x,y);
	}
	
	
	@Override
	public String getAction(Perception perception, Gridworld world,
			IPathFindingHandle pathfinder, MessageService coordinatorMessageService) {
		try {
			// pathfinder befragen
			PathResult pr = pathfinder.getPath(goalPosition,perception.getAgentsPosition(), world, true);
			if(pr.isReachedGoal()){
				// goal is reached
				isReached = true;
				return IMAPCConstants.ACTION_SKIP;
			}else if(pr.getErrortype() != null){
							
				if (pr.getErrortype() == IMAPCConstants.PATH_ERROR){
					return this.getRandomAction();
					// hier gab es wirklich einen Fehler im pathfinder, sollte nie auftreten, wenn doch dann random sicher das beste
				}else if (pr.getErrortype() == IMAPCConstants.PATH_NOTFOUND){
					// The Pathfinder was not able to find a path
					//sollte eigentlich nicht passieren, da das Ziel erreichbar sein sollte, sonst ware es nicht zugeteilt worden
					return this.getRandomAction();
				
				}else if (pr.getErrortype() == IMAPCConstants.PATH_FENCEAHEAD){
					// The Pathfinder indicates there is a fence in the next move,
					//maybe we already voted but the switcher hasn't reached the switching position yet or we didn't vote
					Switch s = world.findSwitchForFence(new Position(pr.getFencex(),pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(),pr.getFencey()),true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
					// If the fence is in the next move, the action is skip
					return pr.getThisStep();
				}else if (pr.getErrortype() == IMAPCConstants.PATH_BLOCKED){
					// The Pathfinder indicates the next move is blocked
					// da is ne kuh oder nen agent ins bild gerannt, // TODO randommove oder lieber skip?
					return this.getRandomAction();
				}
			}else{
				// no error, vote if there are fences ahead
				if(pr.getStepsToNextFence()!=-1){
					//TODO ab welchem Abstand anmelden?
					Position s = world.findSwitchForFence(new Position(pr.getFencex(), pr.getFencey()));
					if(s == null){
						// there is no switch found yet
						coordinatorMessageService.sendSwitchVote(new Position(pr.getFencex(),pr.getFencey()),true, pr.getStepsToNextFence());
					}else{
						// vote for this switch to be opened
						coordinatorMessageService.sendSwitchVote(s, false, pr.getStepsToNextFence());
					}
				}
				//TODO weitere Zauune anmelden
				
				// und es koennte sein, dass keiner erreicht ohne durch Zaun zu gehen
				// koennen wir uns aber spaeter Gedanken drueben machen
				return pr.getThisStep();
			}
		}catch (Exception e){
			e.printStackTrace();
		}
		// hier nur nach exception, //TODO randommove oder lieber skip
		return this.getRandomAction();
	}

	@Override
	public void deserialise(IDeserialiser arg0) throws IOException,
			ClassNotFoundException {
		goalPosition.deserialise(arg0);
		this.isReached = arg0.readBoolean();
	}

	@Override
	public void serialise(ISerialiser arg0) throws IOException {
		goalPosition.serialise(arg0);
		arg0.writeBoolean(this.isReached);
	}

	@Override
	public String toString() {
		return "reach driving position "+goalPosition;
	}

	@Override
	public int getPriority() {
		return 30;
	}

	@Override
	public boolean isAchieved() {
		return isReached;
	}

	@Override
	public boolean equals(Object o) {
		if(o == null){
			return false;
		}
		if(o instanceof GoalReachDrivingPosition){
			GoalReachDrivingPosition other = (GoalReachDrivingPosition) o;
			if (other.goalPosition.equals(goalPosition)){
				return true;
			}
		}
		return false;
	}

	@Override
	public Position getGoalPosition() {
		return goalPosition;
	}

}
