package de.tub.mapc.components.contestagents;

import java.io.IOException;

import org.slf4j.Logger;

import de.dailab.keks.io.IAddress;
import de.dailab.keks.io.IMessage;
import de.jiac.micro.agent.handle.ICommunicationHandle;
import de.jiac.micro.core.IAgent;
import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.Coordination.messages.MessageSwitchVote;
import de.tub.mapc.Coordination.messages.SwitchingAnswer;
import de.tub.mapc.knowledge.Perception;
import de.tub.mapc.knowledge.Position;

public class MessageService {

	private ICommunicationHandle comm;
	private IAddress receiver;
	private Logger _logger;
	
	public MessageService(ICommunicationHandle comm, IAddress receiver) {
		this.comm = comm;
		this.receiver = receiver;
		IAgent agent = AgentScope.getAgentReference();
		_logger = agent.getLogger();
	}

	public void sendPerception(Perception p){
		IMessage message = comm.createMessage();
		message.setContent(p);
		
		try {
			comm.sendMessage(receiver, message);
		} catch (IOException e) {
			_logger.error("failed to share perception with other agents", e);
		}
	}
	
	/**
	 * sends a message coordinator to tell, that the Agent wants a switch to be opened
	 * @param s the position of the switch
	 * @param isFence if the agent does not know the position of the switch,
	 *  the parameter s can describe the position of the fence field and isFence can be set to true
	 */
	public void sendSwitchVote(Position s, boolean isFence, int distance){
		IMessage message = comm.createMessage();
		message.setContent(new MessageSwitchVote(s.getX(),s.getY(),isFence, distance));
		try {
			comm.sendMessage(receiver, message);
		} catch (IOException e) {
			_logger.error("Could not send the switchRequest to the coordinator", e);
		}
	}
	
	/**
	 * sends a message to the coordinator, which describes, how the agent can open the switch
	 * @param x the x coordinate of the switch
	 * @param y thy y coordinate of the switch
	 * @param steps the number of steps the agent needs to reach this switch<br/>
	 * 	if he cannot reach set steps to -1
	 * @param possibleSwitchingPosition the switching position, where the Agent can go to open this switch
	 */
	public void sendSwitchRequestAnswer(int x, int y, int steps, Position possibleSwitchingPosition){
		IMessage message = comm.createMessage();
		message.setContent(new SwitchingAnswer(x, y, steps, possibleSwitchingPosition));
		try {
			comm.sendMessage(receiver, message);
		} catch (IOException e) {
			_logger.error("Could not send the switchRequestAnswer to the coordinator", e);
		}
	}
	
}
