package de.tub.mapc.monitor;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.WindowConstants;

import de.tub.mapc.Coordination.goals.AbstractGoal;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.pathfinding.PathMap;
import de.tub.mapc.worldmodel.Herd;

public class AgentMonitor implements ActionListener {
		
	private static final AgentMonitor INSTANCE = new AgentMonitor();
	
	private JTabbedPane tabPane;
	private ButtonGroup agentButtonGroup;
	private ArrayList<AgentButton> agentButtons;
	private JPanel buttonPane;
	private JTextArea text;
	private JFrame window;
	private JPanel labelPane;
	private JLabel score;
	private JLabel steps;
	private HashMap<Integer,PathMap> agentPathMaps;
	
	private int currStep = 0;
	private int currCows = 0;
	private int agentNumber = 0;
	
	private MapComponent map;
	
	private Gridworld world;
	
	private ArrayList<String> agentIDs;
//	private HashMap<String, ArrayList<Object>> agentMessages;
	private String selectedAgent = "";
	
	private AgentMonitor(){
		
	}
	
	public static AgentMonitor getInstance(){
		return INSTANCE;
	}
	
	public void init(Gridworld world) {
		if(this.world != null) return;
		
		this.world = world;
		
		initWindow();
		
		map.setWorld(world.getGrid());
		agentIDs = new ArrayList<String>();
//		agentMessages = new HashMap<String, ArrayList<Object>>();
		showWindow();
	}
	
	public void reset(){
		this.world = null;
		window.setVisible(false);
	}
	
	private synchronized void createAgentButton(String agentID) {
		agentIDs.add(agentID);
		AgentButton newButton = new AgentButton(agentID); 
		newButton.addActionListener(this);
		agentButtons.add(newButton);
		agentButtonGroup.add(newButton);
		buttonPane.add(newButton);
		if(agentButtons.size() == 1) {
			agentButtons.get(0).setSelected(true);
		}
	}
	
	private synchronized void initWindow(){
		window = new JFrame("MAPC Monitor Team MicroJIAC");
		window.setPreferredSize(new Dimension(world.getGrid()[0].length*8 + 150, world.getGrid().length*8));
		window.setSize(window.getPreferredSize());
		
window.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);

		window.setLayout(new BorderLayout());
	
		tabPane = new JTabbedPane();
		map = new MapComponent();
		agentPathMaps = new HashMap<Integer,PathMap>();
		
		tabPane.addTab("Map", map);
		text = new JTextArea();
		text.setEditable(false);
		tabPane.addTab("Printlns", new JScrollPane(text));
		agentButtonGroup = new ButtonGroup();
		agentButtons = new ArrayList<AgentButton>();
		buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.Y_AXIS));
		
		labelPane= new JPanel();
		labelPane.setLayout(new FlowLayout(FlowLayout.CENTER, 40, 5));
		score=new JLabel("number of cows: "+currCows);
		steps=new JLabel("step/steps: "+currStep+"/"+world.getSteps());
		labelPane.add(score);
		labelPane.add(steps);
		
		window.add(tabPane, BorderLayout.CENTER);
		window.add(buttonPane, BorderLayout.EAST);
		window.add(labelPane, BorderLayout.NORTH);
	}
	
	private void showWindow(){
		window.repaint();
		window.setVisible(true);
	}
	
	public void updateGoalList(AbstractGoal newGoal, String agentID){
		// auf AbstractGoal anpassen
		if(!agentIDs.contains(agentID)){
			createAgentButton(agentID);
		}
		
		// update the Button
		agentButtons.get(agentIDs.indexOf(agentID)).setGoal(newGoal);
		// update the Log
		text.append("Agent " + agentID + " just send his goal: " + newGoal.toString()+ "\n");
		
		if (agentID.equals(selectedAgent)){
			map.setSelectedGoal(newGoal.getGoalPosition());
			//TODO eventuell was noch ueberlegen, um das Goal anzuzeigen
		}
	}
	
//	public void updateWorld(Perception perception, String agentID){
//		if (this.map == null)return;
//		GridField[] newInfo = perception.getGridInfo();
//		world.update(newInfo);
//		map.setWorld(world.getGrid());
//		// update Labels: step/steps, cows in corral
//		if(currStep < perception.getStep()){
//			currCows = perception.getScore();
//			currStep = perception.getStep();
//			score.setText("number of cows: " + currCows);
//			steps.setText("step/steps: " + currStep + "/" + world.getSteps());
//		}
//		
//		if(agentID.equals(selectedAgent)){
//			map.setSelectedPosition(perception.getPosX(), perception.getPosY());
//		}
//	}
	
	public void setWorld(Gridworld w, int step, int cowsInCorroal){
		world = w;
		map.setWorld(world.getGrid());
		currStep = step;
		map.setCurrStep(step);
		currCows = cowsInCorroal;
		score.setText("number of cows: " + currCows);
		steps.setText("step/steps: " + currStep + "/" + world.getSteps());
		//TODO so wird der selectedAgent nicht mehr angezeigt!
	}
	
	public void updateHerds(List<Herd> h){
		map.setHerds(h);	
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		setSelectedAgent(e.getActionCommand());
	}

	private void setSelectedAgent(String agentID) {
		selectedAgent = agentID;
	}
	
	public synchronized int addTab(PathMap map, String agentID) {
		if(agentPathMaps == null)return -1;
		
		agentPathMaps.put(Integer.valueOf(agentNumber),map);
		tabPane.add(String.valueOf(agentNumber), map);
		
		
		agentNumber = agentNumber + 1;
		
		return agentNumber;
	}
	
}
