package de.tub.mapc.pathfinding;

public class State {
	public enum TAGS{ NEW, OPEN, CLOSED, RAISE, LOWER };
	private TAGS tag;
	private State backState;
	private int cost;
	private int previousCost;
	private int pathCost;
	private int posx, posy;
	private String ID;
	
	/**
	 * creates a new State
	 * @param x, position on the grid
	 * @param y, position on the grid
	 */
	public State(int x, int y){
		setTag(TAGS.NEW);
		setBackState(null);
		setCost(0);
		setPreviousCost(0);
		setPathCost(0);
		setPosy(y);
		setPosx(x);
		generateID();
	}
	
	/**
	 * @param iD the iD to set
	 */
	public void generateID() {
		ID = this.posx+"#"+this.posy;
	}

	/**
	 * @return the iD
	 */
	public String getID() {
		return ID;
	}

	public TAGS getTag(){
		return this.tag;
	}
	
	public void setTag(TAGS newTag){
		this.tag = newTag;
	}
	
	public State getBackState(){
		return this.backState;
	}
	
	public void setBackState(State bp){
		this.backState = bp;
	}

	/**
	 * @param previousCost the previousCost to set
	 */
	public void setPreviousCost(int previousCost) {
		this.previousCost = previousCost;
	}

	/**
	 * @return the previousCost
	 */
	public int getPreviousCost() {
		return previousCost;
	}

	/**
	 * @param pathCost the pathCost to set
	 */
	public void setPathCost(int pathCost) {
		this.pathCost = pathCost;
	}

	/**
	 * @return the pathCost
	 */
	public int getPathCost() {
		return pathCost;
	}

	/**
	 * @param posy, y position on the grid
	 */
	public void setPosy(int posy) {
		this.posy = posy;
		generateID();
	}

	/**
	 * @return the y position on the grid
	 */
	public int getPosy() {
		return posy;
	}
	
	/**
	 * @param posx, x position on the grid
	 */
	public void setPosx(int posx) {
		this.posx = posx;
		generateID();
	}

	/**
	 * @return the x position on the grid
	 */
	public int getPosx() {
		return posx;
	}

	/**
	 * @param cost the cost to set
	 */
	public void setCost(int cost) {
		this.cost = cost;
	}

	/**
	 * @return the cost
	 */
	public int getCost() {
		calculateCost();
		return cost;
	}
	
	/**
	 * calculates the cost(X) from the pathCost and previousCost
	 */
	public void calculateCost(){
		if (this.getPathCost() < this.getPreviousCost())
			this.setCost(this.getPathCost());
		else this.setCost(this.getPreviousCost());
	}

	public void reset() {
		setCost(0);
		setPathCost(0);
		setPreviousCost(0);
		setBackState(null);
		setTag(TAGS.NEW);
	}

}
