package de.tub.mapc.worldmodel;

import static de.tub.mapc.IMAPCConstants.ACTION_EAST;
import static de.tub.mapc.IMAPCConstants.ACTION_NORTH;
import static de.tub.mapc.IMAPCConstants.ACTION_NORTHEAST;
import static de.tub.mapc.IMAPCConstants.ACTION_NORTHWEST;
import static de.tub.mapc.IMAPCConstants.ACTION_SOUTH;
import static de.tub.mapc.IMAPCConstants.ACTION_SOUTHEAST;
import static de.tub.mapc.IMAPCConstants.ACTION_SOUTHWEST;
import static de.tub.mapc.IMAPCConstants.ACTION_WEST;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.slf4j.Logger;

import de.jiac.micro.core.scope.AgentScope;
import de.tub.mapc.IMAPCConstants;
import de.tub.mapc.Coordination.Agent;
import de.tub.mapc.Coordination.Voter;
import de.tub.mapc.components.contestagents.PathFindingElement;
import de.tub.mapc.components.contestagents.handle.IPathFindingHandle;
import de.tub.mapc.knowledge.Gridworld;
import de.tub.mapc.knowledge.Position;
import de.tub.mapc.pathfinding.PathResult;

//Die Klasse Herd repraesentiert nur eine Herde
//(bietet die Treibe-Positionen und die Begrenzung 
//und hat die Anzahl der Treiber gespeichert und die ids,
// bietet eine aktualisierungsfunktion, abstand zum corral) 
//@SuppressWarnings("unchecked")
public class Herd implements Voter, Comparable<Herd> {

	private static final int MAX_COWS_PER_DRIVER = 6; //TODO anzahl optimieren
	
	/**
	 * GridWorld
	 */
	private Gridworld world;

	/**
	 * pathfinder
	 */
	IPathFindingHandle pathfinder;

	/**
	 * the GridFields of all cows
	 */
	private List<GridField> cows;

	/**
	 * boundary rectangle (x0,x1,y0,y1): boundary of this herd
	 */
	private int leftBoundary = 9999, rightBoundary = 0, lowerBoundary = 0, upperBoundary = 9999;

	/**
	 * boundary rectangle (cx0,cx1,cy0,cy1): boundary of the corral
	 */
	private int cx0, cx1, cy0, cy1;

	/**
	 * the middle of the herd
	 */
	private int midX, midY;

	/**
	 * the middle of the corral
	 */
	private int cMidX, cMidY;

	/**
	 * the max Number Of Drivers
	 */
	private int maxDrivers;

	/**
	 * list of current drivers
	 */
	private List<Agent> drivers = new ArrayList<Agent>();

	/**
	 * all driving agents, the key is the String representation of the address
	 */
	private HashMap<String, Agent> agents;

	/**
	 * the distance to the corral
	 */
	private double distanceToCorral;

	/**
	 * the array of driving positions
	 */
	private List<Position> drivingPositions = new LinkedList<Position>();

	private List<Point> fences = null;
	private int stepsToNextFence;

	/**
	 * @param cows
	 *            ArrayList of cow positions
	 * @param cX0
	 *            upper x coordinate of the corral
	 * @param cX1
	 *            lower y coordinate of the corral
	 * @param cY0
	 *            upper y coordinate of the corral
	 * @param cY1
	 *            lower y coordinate of the corral
	 * 
	 */
	public Herd(List<GridField> cows, Gridworld w) {
		if(cows.size() == 0){
			System.out.println("WARNUNG, ERZEUGE HERDE OHNE KUEHE!!!!");
			new RuntimeException("STACKTRACE").printStackTrace();
		}
		this.cows = new ArrayList<GridField>();
		this.cows.addAll(cows);
		this.world = w;
		agents = new HashMap<String, Agent>();
		calcBoundaryOfHerd();
		calcMiddleOfTheHerd();
		this.cx0 = world.getCorralX0();
		this.cx1 = world.getCorralX1();
		this.cy0 = world.getCorralY0();
		this.cy1 = world.getCorralY1();
		calcMiddleOfTheCorral();
		calcMaxDriverNo();		
		pathfinder = new PathFindingElement(false);
		calcFieldsToDriveTo();
	}
	
	/**
	 * calculates the boundary of the herd
	 * 
	 * @param cows
	 *            an Arraylist of cows
	 */
	private void calcBoundaryOfHerd() {
		
		// eventuell vorher wieder initialwerte setzen wie oben?
		for (GridField f : this.cows) {
			if (f.getX() < leftBoundary)
				leftBoundary = f.getX();
			if (f.getX() > rightBoundary)
				rightBoundary = f.getX();
			if (f.getY() > lowerBoundary)
				lowerBoundary = f.getY();
			if (f.getY() < upperBoundary)
				upperBoundary = f.getY();
		}
	}

	public double getDistanceToCorral() {
		return distanceToCorral;
	}

	/**
	 * updates the world
	 * 
	 * @param w new Gridworld
	 */
	public void updateWorld(Gridworld w) {
		world = w;
	}
	
	/**
	 * update the herd with a new array of cows
	 * @param stepsBetweenUpdates
	 */
	public void updateHerd(int stepsBetweenUpdates){
		updateCows(stepsBetweenUpdates);
		
		calcBoundaryOfHerd();
		calcMiddleOfTheHerd();
		calcMaxDriverNo();
		calcFieldsToDriveTo();
	}

	
	private void updateCows(int stepsBetweenUpdates){
		// get all cows around the herd (distance to herd = stepsBetweenUpdates+/-1)
		// stepsbetween +/- 1 weil das die Steps sind, in denen nicht geupdatet wird
		// +2 ist das +/-1 und dann noch eins wegen dem < in der schleife. da darf kein <= wegen gridsize wenns am rand is
		List<GridField> newCows = new ArrayList<GridField>();
		GridField gf;
		int leftBoundary = Math.max(0, getUpperLeft().x - stepsBetweenUpdates-1);
		int rightBoundary = Math.min(getLowerRight().x + stepsBetweenUpdates + 2, world.getGsizeX());
		int upperBoundary = Math.max(0, getUpperLeft().y - stepsBetweenUpdates-1);
		int lowerBoundary = Math.min(getLowerRight().y + stepsBetweenUpdates + 2, world.getGsizeY());
		for(int x = leftBoundary; x < rightBoundary; x++){
			for(int y = upperBoundary; y < lowerBoundary; y++){
				gf = world.getGrid()[x][y];
				if(gf.getType() == IMAPCConstants.TYPE_COW)
					newCows.add(gf);
			}
		}
		// if two herds overlap they will be merged in HerdOrganisation
		// and can also be separated there
				
		this.cows.clear();
		this.cows.addAll(newCows);		
		
	}
	
	/**
	 * get the cows
	 * 
	 * @return list of cows
	 */
	public List<GridField> getCows() {
		return this.cows;
	}

	/**
	 * set the cows
	 */
	public void setCows(List<GridField> c) {
		this.cows = c;
	}

	/**
	 * Upper left of the herd
	 * @return Point (x0,y1)
	 */

	public Point getUpperLeft() {
		return new Point(leftBoundary, upperBoundary);
	}

	/**
	 * lower right of the herd
	 * 
	 * @return Point (x1,y0)
	 */
	public Point getLowerRight() {
		return new Point(rightBoundary, lowerBoundary);
	}

	/**
	 * get the size of the herd
	 * 
	 * @return number of cows
	 */
	public int size() {
		return cows.size();
	}
	
	public int getMaxDriverCount() {
		return maxDrivers;
	}
	
	public Gridworld getWorld() {
		return world;
	}

	public int getDriverCount() {
		return drivers.size();
	}

	public Position getMidlle() {
		return new Position(midX, midY);
	}

	/**
	 * returns the pathlength from position p to the middle of the herd
	 * and -1 if the herd is not reachable
	 * @param p
	 * @return
	 */
	public int getPathLengthAsDistance(Position p){
		IPathFindingHandle pf = new PathFindingElement(false);
		// check reachability
		PathResult pr = pf.getPathForHerd(new Position(midX, midY), p, world);
		if(pr.getErrortype() == IMAPCConstants.PATH_NOTFOUND) 
			return -1;
		else return pr.getPathLength();
	}
	
	/**
	 * get the list of driving agents
	 * 
	 * @return list of agents
	 */
	public List<Agent> getDrivers() {
		return drivers;
	}

	/**
	 * add agent to driving agents if the max number of drivers is reached, the
	 * driver won't be added
	 * 
	 * @param a
	 *            Agent
	 * @return true, if agent is added, else false
	 */
	public boolean addDriver(Agent a) {
		if (drivers.size() < maxDrivers && drivers.size() < drivingPositions.size()) {
			if (drivers.add(a)) {
				if (a == null || agents == null || a.getAddress() == null) {
					System.out.println("Herd: addDriver");//TODO logger
				}
				agents.put(a.getAddress().toString(), a);
				return true;
			} else {
				return false;
			}
		} else {
			return false;
		}
	}

	/**
	 * remove agent from driving agents
	 * 
	 * @param a
	 *            Agent
	 * @return true, if agent is removed, else false
	 */
	public boolean removeDriver(Agent a) {
		if (drivers.remove(a)) {
			agents.remove(a.getAddress().toString());
			return true;
		} else {
			return false;
		}
	}

	private void calcMiddleOfTheCorral() {
		cMidX = ((cx0 + cx1) / 2);
		cMidY = ((cy0 + cy1) / 2);
	}

	private void calcMiddleOfTheHerd() {
		midX = ((leftBoundary + rightBoundary) / 2);
		midY = ((lowerBoundary + upperBoundary) / 2);
	}
	
	private void calcMaxDriverNo(){
		this.maxDrivers = (cows.size()/MAX_COWS_PER_DRIVER)+2;
	}

	/**
	 * get the driving position for the Agent
	 * 
	 * Agent mit index 0 bekommt mittlere Position
	 * 
	 * @param a
	 *            the agent
	 * @return the driving position
	 */
	public Position getMyDrivingPos(String id) {
		// TODO nearest position!
		
		if (drivingPositions.size() == 0)
			return null;
		Agent a = agents.get(id);
		
		int indexOfDriver = drivers.indexOf(a);
		int middleOfDrivingPositions = drivingPositions.size()/2;
		
		if(indexOfDriver == 0) {
			if(drivers.size() % 2 == 1)
				return drivingPositions.get(middleOfDrivingPositions);
		}
		
		double distanceBetweenDrivers = ((double)drivingPositions.size() / (double)drivers.size());
		int resultindex = 0;
		
		if(indexOfDriver % 2 == 0) {
			resultindex = middleOfDrivingPositions + (int) Math.ceil(distanceBetweenDrivers * (indexOfDriver/2.0));
			if(resultindex == drivingPositions.size()) resultindex--;
		} else {
			resultindex = middleOfDrivingPositions - (int) Math.ceil(distanceBetweenDrivers * (indexOfDriver/2.0));
		}
		
//		System.out.println(drivers.size()+" drivers on");
//		System.out.println(drivingPositions.size()+" Positions (Middle: "+middleOfDrivingPositions+", distance: "+distanceBetweenDrivers+")");
//		System.out.println(indexOfDriver+" got index "+resultindex);
		
		if (resultindex < 0 && resultindex >= drivingPositions.size())
			return null;
		return drivingPositions.get(resultindex);
		
	}
	
	public List<Position> getDrivingPositions(){
		return this.drivingPositions;
	}
	
	public String toString(){
		return "upperLeft= ("+leftBoundary+","+upperBoundary+") lowerRight= ("+rightBoundary+","+lowerBoundary+") number of cows: "+size();
	}

	public List<Point> getFences() {
		return this.fences;
	}

// FIXME: seems to be unused
//	public Point[] getCorralDimension() {
//		return new Point[]{new Point(cx0,cy0), new Point(cx1, cy1)};
//	}

	/**
	 * Gets orthogonal line (in the form of a few {@code Point}s) to a specified
	 * direction. WARNING: The line will be closest to a cow in the herd - and
	 * can thus be outside the world or on blocked fields! TODO Ist es iO das so
	 * zu machen oder sollten die punkte nur innerhalb der Herde sein? TODO
	 * Welche Groesse sollte der Vektor haben? {@code max width,height}}?!
	 * 
	 */
	public void calcFieldsToDriveTo() {
	    final Logger logger= AgentScope.getAgentReference().getLogger("Herd");
	    
		// get target from pathfinder
		if (cMidX < 0 || midX < 0 || cMidY < 0 || midY < 0) {
			logger.warn("Error in calcFieldsToDriveTo: negative center of herd or corral");
		}
		// get the path from herd to corral
		PathResult pr = pathfinder.getPathForHerd(new Position(cMidX, cMidY), new Position(midX, midY), world);
		// TODO patherror: stehenbleiben oder randomMove?
		if (pr.getErrortype() != null && pr.getErrortype() != IMAPCConstants.PATH_FENCEAHEAD) {
		    if(logger.isWarnEnabled()) {
		        logger.warn("Herd: Couldnt calc path for herd: getPath(+"+cMidX+","+ cMidY+","+ midX+"," +midY);
		        logger.warn("Error: "+pr.getErrortype());
		        logger.warn("Cows: "+cows.size());
		        logger.warn("UL:"+getUpperLeft().toString());
		        logger.warn("LR:"+getLowerRight().toString());
		    }
		    distanceToCorral = -1;
			return;
		}
		distanceToCorral = pr.getPathLength();

		if (pr.isReachedGoal()) {
			return;
		}

		// there are fences in the way to the corral
		if (pr.getStepsToNextFence() != -1) {
			fences = pr.getFences();
			stepsToNextFence = pr.getStepsToNextFence();
		}

		// get the position of the next step
		Position fieldForDirection = pr.getThreeStepsPosition();
		if(fieldForDirection == null)
			fieldForDirection = getTargetPosition(pr.getThisStep());
		int nextStepX = fieldForDirection.getX();
		int nextStepY = fieldForDirection.getY();

		// line in direction to the next step (through the current middle of the herd and the middle in the next step)
		// line equation 
		// y = directionM * x + directionB
		double directionM;
		double directionB;
		if (nextStepX == midX) { // black hole avoider
			// the direction is vertical (that is no mathematical function) 
			directionM = 99999999;
			directionB = (nextStepY - midY) * (-midX) / (0.0000001) + midY;
		} else {
			directionM = (nextStepY - midY) / (nextStepX - midX);
			directionB = -(directionM * midX) + midY;
		}

		/*
		 * intersection with normal
		 * intersectX and intersectY are the positions, where the direction-line
		 * crosses the boundary of the herd (in the back of the herd)
		 * there the vector with the driving-positions should have his middle
		 */
		double intersectX;
		double intersectY;
		if (Math.abs(directionM) < 1) { // angle < 45 go for x
			// Bewegungsrichtung entlang der x-Achse
			if (nextStepX > midX) {
				// Bewegung nach rechts
				intersectX = leftBoundary;
			} else {
				// Bewegung nach links
				intersectX = rightBoundary;
			}
			intersectY = directionM * intersectX + directionB;
		} else { // go for y
			// Bewegungsrichtung entlang der y-Achse
			if (nextStepY < midY) {
				// Bewegung nach oben
				intersectY = lowerBoundary;
			} else {
				// Bewegung nach unten
				intersectY = upperBoundary;
			}

			if (directionM == 99999999) { // black hole avoider
				intersectX = midX;
			} else {
				intersectX = (intersectY - directionB) / directionM;
			}
		}

		// normal
		double normalM;
		if (directionM == 0) { // black hole avoider
			normalM = 99999999;
		} else {
			normalM = -1 / directionM;
		}
		double normalB = intersectY - (intersectX * normalM);

		// now just fill the array list :)
		int width = rightBoundary - leftBoundary;
		int height = lowerBoundary - upperBoundary;
		List<Position> positions = new ArrayList<Position>();
		if (Math.abs(normalM) < 1) { // again: go for x if angle to x axis < 45
			// Waagerechte Linie mit Treibepositionen
			
			// end - start = Laenge der Treibepositionen-Linie
			int start = Math.max((int) Math.round(intersectX - ((width+height) / 4)), 0);
			int end = Math.min((int) Math.round(intersectX + ((width+height) / 4)), world.getGrid()[0].length);

			// Positionen erzeugen: x wird einfach erhoeht, y wird berechnet
			for (int targetX = start; targetX <= end; targetX++) {
				positions.add(new Position(targetX, (int) Math.round(normalM * targetX + normalB)));
			}
		} else {
			// Senkrechte Linie mit Treibepositionen 

			// end - start = Laenge der Treibepositionen-Linie
			int start = Math.max((int) Math.round(intersectY - ((width+height) / 4)), 0);
			int end = Math.min((int) Math.round(intersectY + ((width+height) / 4)), world.getGrid().length);

			// Positionen erzeugen: x wird berechnet, y wird einfach erhoeht
			for (int targetY = start; targetY <= end; targetY++) {
				positions.add(new Position((int) Math.round((targetY - normalB) / normalM),	targetY));
			}
		}

		drivingPositions.clear();
		drivingPositions.addAll(positions);
	}

	/**
	 * calculates the herd's target position for a given direction
	 * 
	 * @param direction
	 *            the direction
	 * @return the position
	 */
	private Position getTargetPosition(String direction) {
		// get target position
		int tX = midX;
		int tY = midY;
		if (direction.equals(ACTION_NORTH))
			tY += -1;
		if (direction.equals(ACTION_NORTHWEST)) {
			tY += -1;
			tX += -1;
		}
		if (direction.equals(ACTION_NORTHEAST)) {
			tY += -1;
			tX += 1;
		}
		if (direction.equals(ACTION_SOUTH))
			tY += 1;
		if (direction.equals(ACTION_SOUTHWEST)) {
			tY += 1;
			tX += -1;
		}
		if (direction.equals(ACTION_SOUTHEAST)) {
			tY += 1;
			tX += 1;
		}
		if (direction.equals(ACTION_WEST))
			tX -= 1;
		if (direction.equals(ACTION_EAST))
			tX += 1;
		return new Position(tX, tY);
	}

	@Override
	public int compareTo(Herd h) {
		if (this.distanceToCorral < h.getDistanceToCorral())
			return -1;
		else if (this.distanceToCorral > h.getDistanceToCorral())
			return 1;
		else
			return 0;
	}

	/**
	 * 
	 * @return true, if the whole herd is inside the corral
	 */
	public boolean insideCorral() {
		if (getUpperLeft().x >= cx0 && getLowerRight().x <= cx1 && getUpperLeft().y >= cy0 && getLowerRight().y <= cy1){
			// entire herd is inside the corral
			return true;
		}
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Collection<Agent> getAgents() {
		if(drivers.size()>1)
			return drivers;
		//if there is only one driver this driver should not leave the herd
		else return new ArrayList<Agent>();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getPrio() {
		// TODO welche Prio? Ich hab jetzt erstmal ne hoehere als die vom Goal DriveHerd genommen
		return 50;
	}

	public int getStepsToNextFence() {
		return stepsToNextFence;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int getValidSteps() {
		return Math.max(rightBoundary-leftBoundary, lowerBoundary-upperBoundary);
	}

}
